{-# LANGUAGE NamedFieldPuns, RecordWildCards #-}

import Rubik.Cube
import Rubik.Misc
import qualified Rubik.Solver.Optimal as Optimal
import qualified Rubik.Solver.TwoPhase as TwoPhase
import qualified Rubik.Tables.Internal as Option

import Control.Exception
import Control.Monad

import Data.Time.Clock

import Data.Char
import Data.Monoid

import Numeric ( showFFloat )

import Options.Applicative hiding ( value )
import qualified Options.Applicative as Opt

import System.Exit
import System.IO.Error

import Control.Parallel.Strategies(Eval,rpar,runEval,using,parListChunk,rdeepseq)

type Solver = Cube -> Move

data Parameters = Parameters {
    verbose :: Bool,
    solve :: Solver,
    tsPath :: Maybe FilePath,
    precompute :: Bool,
    overwrite :: Bool,
    noFiles :: Bool,
    strict :: Bool,
    debug :: Bool
  }

optparse :: Parser Parameters
optparse = Parameters
  <$> switch ( long "verbose" <> short 'v'
        <> help "Print time taken to solve every cube" )
  <*> flag TwoPhase.solve Optimal.solve ( long "optimal"
        <> help "Use optimal solver (experimental)" )
  <*> (optional . strOption) ( long "ts-dir" <> short 'd'
        <> metavar "DIR"
        <> help "Location of precomputed tables" )
  <*> switch ( long "precompute" <> short 'p'
        <> help "Precompute and store tables \
                \(do enable this at the first invocation)" )
  <*> switch ( long "overwrite"
        <> help "Recompute and overwrite tables even when they exist already" )
  <*> switch ( long "no-files"
        <> help "Do not read or write any files \
                \(recompute tables for this session)" )
  <*> switch ( long "strict"
        <> help "Force loading tables before doing anything else" )
  <*> switch ( long "debug" )


parallel_map::(a->b)->[a]->Eval [b]
parallel_map func [] =return []
parallel_map func (x:xs) = do
  b<- rpar (func x)
  bs <- parallel_map func xs
  return (b:bs)


main :: IO ()
main = do
    p <- execParser $ info (helper <*> optparse) briefDesc
    setOptions p
    file <- readFile "test-40k.txt"
    let cubes = lines file
        --solutions = runEval(parallel_map (faceletList2 p) cubes)
        solutions = map (faceletList2 p) cubes `using` parListChunk 1000 rdeepseq
        --solutions = map (faceletList2 p) cubes
    print(solutions)
    return ()

setOptions :: Parameters -> IO ()
setOptions Parameters{..} = do
  mapM_ Option.setTsPath tsPath
  Option.setPrecompute precompute
  Option.setOverwrite overwrite
  Option.setNoFiles noFiles
  Option.setDebug debug
  when strict . void $ evaluate
    (solve . either undefined moveToCube . stringToMove $ "ulfrbd")

-- A sequence of moves, e.g., "URF".
moveSequence s = putStrLn $
  case stringToMove s of
    Left c -> "Unexpected '" ++ [c] ++ "'."
    Right ms -> stringOfCubeColors . moveToCube . reduceMove $ ms

faceletList2 p s = case readCube (filter (not . isSpace) s) of
    Left err -> show err
    Right cube -> justSolve2 p cube

readCube s
  = case colorFacelets'' s of
      Nothing -> Left "Expected string of length 54 of a set of (any) 6 \
                      \characters. Centers must be distinct."
      Just colors ->
        case colorFaceletsToCube colors of
          Left fs ->
            Left $ "Facelets " ++ show fs
                ++ " (" ++ show (map (s !!) fs) ++ ") \
                   \do not match any regular cubie."
          Right Nothing ->
            Left "Not a permutation of cubies \
                 \(a cubie is absent, and a cubie occurs twice)."
          Right (Just c) | solvable c -> Right c
          _ -> Left "Unsolvable cube."

justSolve2 :: Parameters -> Cube -> String
justSolve2 p c = moveToString (solve p c)

unlessQuiet' :: IO () -> Parameters -> IO ()
unlessQuiet' a = unlessQuiet (const a) ()

-- Strict in its second argument
unlessQuiet :: (a -> IO ()) -> a -> Parameters -> IO ()
unlessQuiet f a p = evaluate a >> when (verbose p) (f a)

clock :: IO a -> IO Double
clock a = do
  t <- getCurrentTime
  a
  t' <- getCurrentTime
  return (diffTimeToSeconds (diffUTCTime t' t))
  where
    diffTimeToSeconds = fromRational . toRational

listSeq' :: [a] -> [a]
listSeq' s = s `listSeq` s

vPutStrLn :: String -> Parameters -> IO ()
vPutStrLn s = unlessQuiet putStrLn (listSeq' s)

vPutStr :: String -> Parameters -> IO ()
vPutStr s = unlessQuiet putStrLn (listSeq' s)
